{% extends "bukuserver/base.html" %}

{% block content %}
<div class="container">
  <div style="padding: 40px 15px; text-align: center;">
    <h1>BUKU</h1>
    <p class="lead">Bookmark manager like a text-based mini-web</p>
    <p>
      <a class="btn btn-lg btn-success" href="{{url_for('bookmarks-html')}}" role="button">Bookmarks</a>
      <a class="btn btn-lg btn-success" href="{{url_for('get_tags-html')}}" role="button">Tags</a>
      <a class="btn btn-lg btn-success" href="{{url_for('statistic.index')}}" role="button">Statistics</a>
    </p>
    <div class=" col-md-4 col-md-offset-4">
      {{show_bookmarks_search_form(checkbox_class="text-left col-sm-offset-2")}}
    </div>
  </div>
</div>
{% endblock %}
